import datetime
import json
import random
import time
import click

import spade


class PullAgent(spade.agent.Agent):

    async def setup(self):
        self.value = random.randint(1, 1000)
        self.value_time = time.time()
        self.values = {self.name: {"value": self.value, "timestamp": self.value_time}}

        start_at = datetime.datetime.now() + datetime.timedelta(seconds=5)
        self.add_behaviour(self.PullBehaviour(period=2, start_at=start_at))
        template = spade.template.Template(metadata={"performative": "PULL"})
        self.add_behaviour(self.RecvPullBehaviour(), template)
        template = spade.template.Template(metadata={"performative": "REPLY"})
        self.add_behaviour(self.RecvReplyBehaviour(), template)

        print("{} ready.".format(self.name))

    def add_value(self, jid, data):
        if jid in self.values:
            if self.values[jid]["timestamp"] < data["timestamp"]:
                self.values[jid] = data
        else:
            self.values[jid] = data

    def add_contacts(self, contact_list):
        self.contacts = [c.jid for c in contact_list if c.jid != self.jid]
        # print("Adding contacts to {}: {}".format(self.name, self.contacts))
        self.length = len(self.contacts)

    class PullBehaviour(spade.behaviour.PeriodicBehaviour):

        async def run(self):
            k = random.randint(1, self.agent.length)
            print("{} period with k={}!".format(self.agent.name, k))
            random_contacts = random.sample(self.agent.contacts, k)
            print("{} sending to {}".format(self.agent.name, [x.localpart for x in random_contacts]))

            for jid in random_contacts:
                timestamp = self.agent.values[jid.localpart]["timestamp"] if jid.localpart in self.agent.values else None
                body = json.dumps({"timestamp": timestamp})
                msg = spade.message.Message(to=str(jid), body=body, metadata={"performative": "PULL"})
                await self.send(msg)

    class RecvPullBehaviour(spade.behaviour.CyclicBehaviour):
        async def run(self):
            msg = await self.receive(timeout=2)
            if msg:
                body = json.loads(msg.body)
                if body["timestamp"] is None or self.agent.value_time > body["timestamp"]:
                    body = json.dumps({"value": self.agent.value, "timestamp": self.agent.value_time})
                    msg = spade.message.Message(to=str(msg.sender), body=body, metadata={"performative": "REPLY"})
                    await self.send(msg)

    class RecvReplyBehaviour(spade.behaviour.CyclicBehaviour):
        async def run(self):
            msg = await self.receive(timeout=2)
            if msg:
                body = json.loads(msg.body)
                self.agent.add_value(msg.sender.localpart, body)

                print("[{}] <{}>".format(self.agent.name, [x["value"] for x in self.agent.values.values()]))


@click.command()
@click.option('--count', default=10, help='Number of agents.')
def main(count):
    agents = []
    print("Creating {} agents...".format(count))
    for x in range(1, count + 1):
        print("Creating agent {}...".format(x))
        agents.append(PullAgent("pull_agent_17264_{}@localhost".format(x), "test"))

    for ag in agents:
        ag.add_contacts(agents)
        ag.values = {}

    for ag in agents:
        ag.start()

    while True:
        try:
            time.sleep(1)
            if all([len(ag.values) == count for ag in agents]):
                print("Gossip done.")
                break
        except KeyboardInterrupt:
            break

    for ag in agents:
        ag.stop()
    print("Agents finished")


if __name__ == '__main__':
    main()
